/* fft */ 
#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include "fft.h"
#include "config.h"

void ModuleFFT (float *, float *, short);

   /********************************************************************
   * FUNCTION: ModuleFFT
   * PURPOSE: modul of fast Fourier transform
   * INPUTS:  
   *	pfSig[0:swN-1] - signal frame 
   *    swLog2N - log base 2 of no. of points of fft, max 13 (N=4096)
   *
   * OUTPUT: 
   *	pfModSpec[0:swN-1] - modul of fft spectrum
   *
   * RETURN: 
   *	none
   *
   * DESCRIPTION: fft is computed by the standard butterfly algorithm.
   *	The decimation is done in frequency, with the signal in natural
   *	indexation and the spectrum in reverse bit indexation. This guaramtees
   *	the natural order of coefficients for butterflies.
   * KEYWORDS:
   *********************************************************************/
   
void ModuleFFT (float *pfSig, float *pfModSpec, short swLog2N)
{
   static short swFirstPass=1;
   static float *pfReal, *pfImag, **ppfWReal, **ppfWImag;
   static short *pswRevBitIndex, swN, swNHalf;
//   static short pswPowers2[]={1,2,4,8,16,32,64,128,256,512,1024,2048,4096};
   short i,j,k, swBitDirect, swBitInverse, swButterflyStep, swStage;
   short swButterfliesDone, swFirstButterflyInSection, swStep;
   float fAngle, fNewAReal, fNewAImag, fNewBReal, fNewBImag;
   
   if (swFirstPass) {
      
      /* --- number of pts --- */
      
      swN = 1 << swLog2N;
      swNHalf = swN >> 1;
      
      /* --- allocations --- */
      
      /* rev. bit. indexing */
      pswRevBitIndex = (short*) malloc (swN * sizeof(short));    
      /* real. part of sig/spec */
      pfReal = (float*) malloc (swN * sizeof(float));    
      /* imag. part of sig/spec */
      pfImag = (float*) malloc (swN * sizeof(float));   
      /* coeffs. for butterflies */ 
      ppfWReal = (float**) malloc (swLog2N * sizeof(float*)); 
      ppfWImag = (float**) malloc (swLog2N * sizeof(float*)); 
      for (i=0; i<swLog2N; i++) {
	ppfWReal[i] = (float*) malloc (swNHalf * sizeof(float));   
	ppfWImag[i] = (float*) malloc (swNHalf * sizeof(float));   
      }
      
      /* ------------------------------------------------- */
      /* --- must fill the reverse bit indexing buffer --- */
      /* ------------------------------------------------- */
      
      for (i=0; i<swN; i++)
	pswRevBitIndex[i]=0;
      
      swBitDirect=1;
      swBitInverse=swNHalf;
      for (i=0; i<swLog2N; i++) {
	 for (j=0; j<swN; j++)
	    if (j & swBitDirect)
	       pswRevBitIndex[j] += swBitInverse;
	 swBitDirect = swBitDirect << 1;
	 swBitInverse = swBitInverse >> 1;
      }
      
      /* ------------------------------------------------------- */
      /* --- must also fill the coefficients for butterflies --- */
      /* ------------------------------------------------------- */
      
      /* --- the first stage has W^0, W^1 ... W^(N/2-1) */
      
      for (i=0; i<swNHalf; i++) {
	 fAngle = -2*PI*(float)i / (float)swN;
         ppfWReal[0][i] = cos (fAngle);
         ppfWImag[0][i] = sin (fAngle);
      }
      
      /* --- next stages are done by the decimation of the 1st --- */
      
      swStep = 2;
      for (swStage = 1; swStage < swLog2N; swStage++) {
	 for (i=0, j=0; i<swNHalf; i+=swStep, j++) {  /* j indexes the downsampled
							 version, i the original */
             ppfWReal[swStage][j] = ppfWReal[0][i];
             ppfWImag[swStage][j] = ppfWImag[0][i];
	 }
	 swStep = swStep << 1; 		/* more coarse downsampling for nex section*/
      }
      
      /* --- that's all for 1st pass ---*/
      
      swFirstPass = 0;
   }
   
   /* --------------------------------------------------------------------- */
   /* --- the function body begins here                                 --- */
   /* --------------------------------------------------------------------- */

   /* --- first, copy the signal (real only) and zero the imag. part ---- */
   
   for (i=0; i<swN; i++) {
      pfReal[i] = pfSig[i];
      pfImag[i] = 0.0;
      printf ("%f ",pfReal[i]); 
   }
   printf ("\n"); 
   
   /* ------------------------------- */
   /* --- then go on to the FFT ! --- */
   /* ------------------------------- */
   
   swButterflyStep = swNHalf;
   
   for (swStage=0; swStage < swLog2N; swStage++) {
      
      swButterfliesDone = 0;
      swFirstButterflyInSection = 0;

      while (swButterfliesDone < swNHalf) {	/* must do some more */
	 
	 /* --- i - indexing of the upper part of butterfly 
	        j - indexing of the lower part of butterfly
		k - indexing of coefficient (power of W)    --- */
	 
	 for (i=swFirstButterflyInSection, 
	      j=swFirstButterflyInSection+swButterflyStep,
	      k=0;
	      i<swFirstButterflyInSection+swButterflyStep;
	      i++, j++, k++) {
	    
	    /* --- upper part of the butt. without multipl. --- */
	    fNewAReal = pfReal[i]+pfReal[j];
	    fNewAImag = pfImag[i]+pfImag[j];
	    
	    /* --- lower part with multipl. --- */
	    fNewBReal = ( pfReal[i] - pfReal[j] ) * ppfWReal[swStage][k]
	                - ( pfImag[i] - pfImag[j] ) * ppfWImag[swStage][k];
	    fNewBImag = ( pfReal[i] - pfReal[j] ) * ppfWImag[swStage][k]
	                + ( pfImag[i] - pfImag[j] ) * ppfWReal[swStage][k];
	    
	    /* --- store --- */
	    pfReal[i] = fNewAReal;
	    pfImag[i] = fNewAImag;

	    pfReal[j] = fNewBReal;
	    pfImag[j] = fNewBImag;
	    
	    /* --- butterfly ready --- */
	    swButterfliesDone++;
	 }
	 
	 /* ----- that's all for this section of butterflies --- */
	 
	 swFirstButterflyInSection += swButterflyStep << 1;
	 
      }
      
      /* ----- after each  stage update of step --- */
      
      swButterflyStep = swButterflyStep >> 1;	/* will be 2 times smaller */
   }
   
   /* ----------------------------------------------------- */
   /* at the end, take the module (with rev. bit. indexing) */
   /* ----------------------------------------------------- */
   
   for (i=0; i<swN; i++) {
      j=pswRevBitIndex[i];
      pfModSpec[i] = sqrt ( pfReal[j] * pfReal[j] + pfImag[j] * pfImag[j]);
   }
}
